import java.util.concurrent.ConcurrentSkipListSet;

import javax.transaction.TransactionManager;

import org.infinispan.Cache;
import org.infinispan.configuration.cache.CacheMode;
import org.infinispan.configuration.cache.Configuration;
import org.infinispan.configuration.cache.ConfigurationBuilder;
import org.infinispan.configuration.cache.VersioningScheme;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.transaction.LockingMode;
import org.infinispan.transaction.TransactionMode;
import org.infinispan.transaction.lookup.DummyTransactionManagerLookup;
import org.infinispan.util.concurrent.IsolationLevel;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class WriteSkewTest
{
    private Cache<String, Integer> cache;

    @Before
    public void createCacheManagers() throws Throwable
    {
        final Configuration config = new ConfigurationBuilder().clustering()
                .cacheMode(CacheMode.LOCAL).transaction()
                .transactionManagerLookup(new DummyTransactionManagerLookup())
                .transactionMode(TransactionMode.TRANSACTIONAL)
                .lockingMode(LockingMode.OPTIMISTIC).syncCommitPhase(true)
                .locking().isolationLevel(IsolationLevel.REPEATABLE_READ)
                .writeSkewCheck(true).versioning().enable()
                .scheme(VersioningScheme.SIMPLE).build();

        final DefaultCacheManager defaultCacheManager = new DefaultCacheManager(
                config);

        cache = defaultCacheManager.getCache("cache");

        cache.start();
    }

    @Test
    public void testSharedCounter() throws Exception
    {
        int counterMaxValue = 1000;
        Cache<String, Integer> c1 = cache;

        // initialize the counter
        c1.put("counter", 0);

        // check if the counter is initialized in all caches
        Assert.assertTrue(c1.get("counter") == 0);

        // this will keep the values put by both threads. any duplicate value
        // will be detected because of the
        // return value of add() method
        ConcurrentSkipListSet<Integer> uniqueValuesIncremented = new ConcurrentSkipListSet<Integer>();

        // create both threads (simulate a node)
        IncrementCounterThread ict1 = new IncrementCounterThread("Node-1", c1,
                uniqueValuesIncremented, counterMaxValue);
        IncrementCounterThread ict2 = new IncrementCounterThread("Node-1", c1,
                uniqueValuesIncremented, counterMaxValue);

        // start and wait to finish
        ict1.start();
        ict2.start();

        ict1.join();
        ict2.join();

        // check if all caches obtains the counter_max_values
        Assert.assertTrue(c1.get("counter") >= counterMaxValue);

        // check is any duplicate value is detected
        Assert.assertTrue(ict1.result);
        Assert.assertTrue(ict2.result);
    }

    private static class IncrementCounterThread extends Thread
    {
        private Cache<String, Integer> cache;

        private ConcurrentSkipListSet<Integer> uniqueValuesSet;

        private TransactionManager transactionManager;

        private int lastValue;

        private boolean result = true;

        private int counterMaxValue;

        public IncrementCounterThread(String name,
                Cache<String, Integer> cache,
                ConcurrentSkipListSet<Integer> uniqueValuesSet,
                int counterMaxValue)
        {
            super(name);
            this.cache = cache;
            this.transactionManager = cache.getAdvancedCache()
                    .getTransactionManager();
            this.uniqueValuesSet = uniqueValuesSet;
            this.lastValue = 0;
            this.counterMaxValue = counterMaxValue;
        }

        @Override
        public void run()
        {
            while (lastValue < counterMaxValue)
            {
                try
                {
                    try
                    {
                        // start transaction, get the counter value, increment
                        // and put it again
                        // check for duplicates in case of success
                        transactionManager.begin();

                        Integer value = cache.get("counter");
                        value = value + 1;
                        lastValue = value;

                        cache.put("counter", value);

                        transactionManager.commit();

                        result = result & uniqueValuesSet.add(value);
                    }
                    catch (Throwable t)
                    {
                        // lets rollback
                        transactionManager.rollback();
                    }
                }
                catch (Throwable t)
                {
                    // the only possible exception is thrown by the rollback.
                    // just ignore it
                }
            }
        }
    }
}
